@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2014-2021 Canonical Ltd.
- * Copyright (C) 2021 Colin Ian King.
+ * Copyright (C) 2021-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -139,7 +139,8 @@ typedef enum {
 	STAT_PTRC,		/* Ptrace */
 	STAT_UID,		/* UID change */
 	STAT_SID,		/* SID change */
-	STAT_LAST		/* Always last sentinal */
+	STAT_LAST,		/* Always last sentinal */
+	STAT_NONE
 } event_t;
 
 typedef struct proc_stats {
@@ -174,8 +175,8 @@ static const ev_map_t ev_map[] = {
 	{ "ptrce","Ptrace",	OPT_EV_PTRC,	STAT_PTRC },
 	{ "uid",  "Uid",	OPT_EV_UID,	STAT_UID  },
 	{ "sid",  "Sid",	OPT_EV_SID,	STAT_SID  },
-	{ "all",  "",		OPT_EV_ALL,	0 },
-	{ NULL,	  NULL, 	0,		0 }
+	{ "all",  "",		OPT_EV_ALL,	STAT_NONE },
+	{ NULL,	  NULL, 	0,		STAT_NONE }
 };
 
 #define KERN_TASK_INFO(str)	{ str, sizeof(str) - 1 }
@@ -189,6 +190,7 @@ static tty_name_info_t *tty_name_info[MAX_TTYS];/* TTY dev to name hash table */
 static unsigned int opt_flags = OPT_CMD_LONG;	/* Default option */
 static int row = 0;				/* tty row number */
 static long int opt_duration = -1;		/* duration, < 0 means run forever */
+static pid_t opt_pgrp = -1;
 
 static char unknown[] = "<unknown>";
 
@@ -299,7 +301,7 @@ static char *proc_comm_dup(const char *str)
  */
 static const char *get_proc_self_stat_field(const char *buf, const int num)
 {
-	const char *ptr = buf, *comm_end;
+	const char *ptr, *comm_end;
 	int n;
 
 	if (num < 1 || !buf || !*buf)
@@ -321,7 +323,7 @@ static const char *get_proc_self_stat_field(const char *buf, const int num)
 
 	ptr = comm_end;
 	while (*ptr) {
-		while (*ptr && *ptr == ' ')
+		while (*ptr == ' ')
 			ptr++;
 		n--;
 		if (n <= 0)
@@ -1148,8 +1150,9 @@ static void free_proc_comm(char *comm)
 static char *proc_cmdline(const pid_t pid)
 {
 	int fd;
+	const size_t buffer_sz = 65536;
 	ssize_t ret;
-	char buffer[4096];
+	char *buffer, *tmp;
 
 	if (pid == 0)
 		return proc_comm_dup("[swapper]");
@@ -1157,12 +1160,19 @@ static char *proc_cmdline(const pid_t pid)
 	if (opt_flags & OPT_COMM)
 		return proc_comm(pid);
 
-	(void)snprintf(buffer, sizeof(buffer), "/proc/%d/cmdline", pid);
-	if ((fd = open(buffer, O_RDONLY)) < 0)
+	buffer = malloc(buffer_sz);
+	if (!buffer)
 		return proc_comm(pid);
 
-	(void)memset(buffer, 0, sizeof(buffer));
-	if ((ret = read(fd, buffer, sizeof(buffer) - 1)) <= 0) {
+	(void)snprintf(buffer, buffer_sz, "/proc/%d/cmdline", pid);
+	if ((fd = open(buffer, O_RDONLY)) < 0) {
+		free(buffer);
+		return proc_comm(pid);
+	}
+
+	(void)memset(buffer, 0, buffer_sz);
+	if ((ret = read(fd, buffer, buffer_sz - 1)) <= 0) {
+		free(buffer);
 		(void)close(fd);
 		return proc_comm(pid);
 	}
@@ -1177,19 +1187,21 @@ static char *proc_cmdline(const pid_t pid)
 
 		for (ptr = buffer; ptr < buffer + ret; ptr++) {
 			if (*ptr == '\0') {
-				if (*(ptr + 1) == '\0')
-					break;
 				*ptr = ' ';
 			}
 		}
+		*ptr = '\0';
 	}
 
 	proc_name_clean(buffer, ret);
 
-	if (opt_flags & OPT_CMD_DIRNAME_STRIP)
-		return proc_comm_dup(basename(buffer));
-
-	return proc_comm_dup(buffer);
+	if (opt_flags & OPT_CMD_DIRNAME_STRIP) {
+		tmp = proc_comm_dup(basename(buffer));
+	} else {
+		tmp = proc_comm_dup(buffer);
+	}
+	free(buffer);
+	return tmp;
 }
 
 /*
@@ -1559,7 +1571,7 @@ static int monitor(const int sock)
 			char when[10];
 			time_t now;
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
-			pid_t pid, ppid;
+			pid_t pid, ppid, pgrp;
 			bool is_thread;
 #endif
 
@@ -1600,6 +1612,9 @@ static int monitor(const int sock)
 			case PROC_EVENT_FORK:
 				ppid = get_parent_pid(proc_ev->event_data.fork.child_pid, &is_thread);
 				pid = proc_ev->event_data.fork.child_pid;
+				pgrp = getpgid(pid);
+				if ((opt_pgrp >= 0) && (opt_pgrp != pgrp))
+					break;
 				proc_stats_account(proc_ev->event_data.fork.parent_pid,
 					is_thread ? STAT_CLNE : STAT_FORK);
 				if (gettimeofday(&tv, NULL) < 0)
@@ -1641,6 +1656,9 @@ static int monitor(const int sock)
 			case PROC_EVENT_EXEC:
 				proc_stats_account(proc_ev->event_data.exec.process_pid, STAT_EXEC);
 				pid = proc_ev->event_data.exec.process_pid;
+				pgrp = getpgid(pid);
+				if ((opt_pgrp >= 0) && (opt_pgrp != pgrp))
+					break;
 				info1 = proc_info_update(pid);
 				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_EXEC)) {
 					row_increment();
@@ -1659,6 +1677,9 @@ static int monitor(const int sock)
 				proc_stats_account(proc_ev->event_data.exit.process_pid, STAT_EXIT);
 				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_EXIT)) {
 					pid = proc_ev->event_data.exit.process_pid;
+					pgrp = getpgid(pid);
+					if ((opt_pgrp >= 0) && (opt_pgrp != pgrp))
+						break;
 					info1 = proc_info_get(pid);
 					if (info1->start.tv_sec) {
 						double d1, d2;
@@ -1692,6 +1713,9 @@ static int monitor(const int sock)
 				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_UID)) {
 					row_increment();
 					pid = proc_ev->event_data.exec.process_pid;
+					pgrp = getpgid(pid);
+					if ((opt_pgrp >= 0) && (opt_pgrp != pgrp))
+						break;
 					if (proc_ev->what == PROC_EVENT_UID) {
 						(void)printf("%s uid   %*d %s%s%6s %8s %s%s%s\n",
 							when,
@@ -1722,6 +1746,9 @@ static int monitor(const int sock)
 				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_UID)) {
 					row_increment();
 					pid = proc_ev->event_data.exec.process_pid;
+					pgrp = getpgid(pid);
+					if ((opt_pgrp >= 0) && (opt_pgrp != pgrp))
+						break;
 					(void)printf("%s sid   %*d %s%s%6d %8s %s%s%s\n",
 						when,
 						pid_size, pid,
@@ -1740,6 +1767,9 @@ static int monitor(const int sock)
 				proc_stats_account(proc_ev->event_data.coredump.process_pid, STAT_CORE);
 				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_CORE)) {
 					pid = proc_ev->event_data.coredump.process_pid;
+					pgrp = getpgid(pid);
+					if ((opt_pgrp >= 0) && (opt_pgrp != pgrp))
+						break;
 					info1 = proc_info_get(pid);
 					row_increment();
 					(void)printf("%s core  %*d %s%s       %8s %s%s%s\n",
@@ -1766,6 +1796,9 @@ static int monitor(const int sock)
 #else
 					pid = proc_ev->event_data.ptrace.process_pid;
 #endif
+					pgrp = getpgid(pid);
+					if ((opt_pgrp >= 0) && (opt_pgrp != pgrp))
+						break;
 					info1 = proc_info_get(pid);
 					row_increment();
 					(void)printf("%s ptrce %*d %s%s%6s %8s %s%s%s\n",
@@ -1786,6 +1819,9 @@ static int monitor(const int sock)
 				proc_stats_account(proc_ev->event_data.comm.process_pid, STAT_COMM);
 				if (!(opt_flags & OPT_QUIET) && (opt_flags & OPT_EV_COMM)) {
 					pid = proc_ev->event_data.comm.process_pid;
+					pgrp = getpgid(pid);
+					if ((opt_pgrp >= 0) && (opt_pgrp != pgrp))
+						break;
 					info1 = proc_info_get(pid);
 					comm = proc_cmdline(pid);
 					row_increment();
@@ -1875,7 +1911,8 @@ int main(int argc, char * const argv[])
 	struct sigaction new_action;
 
 	for (;;) {
-		const int c = getopt(argc, argv, "cdD:e:EghlrsSqxX");
+		const int c = getopt(argc, argv, "cdD:e:EghlrsSp:qxX");
+
 		if (c == -1)
 			break;
 		switch (c) {
@@ -1888,7 +1925,7 @@ int main(int argc, char * const argv[])
 		case 'D':
 			opt_duration = strtol(optarg, NULL, 10);
 			if (opt_duration <= 0) {
-				(void)fprintf(stderr, "Illegal duration.\n");
+				(void)fprintf(stderr, "Illegal duration '%s'.\n", optarg);
 				exit(EXIT_FAILURE);
 			}
 			break;
@@ -1915,6 +1952,13 @@ int main(int argc, char * const argv[])
 		case 'S':
 			opt_flags |= OPT_STATS;
 			break;
+		case 'p':
+			opt_pgrp = strtol(optarg, NULL, 10);
+			if (opt_pgrp < 0) {
+				(void)fprintf(stderr, "Illegal pgroup '%s'.\n", optarg);
+				exit(EXIT_FAILURE);
+			}
+			break;
 		case 'q':
 			opt_flags |= OPT_QUIET;
 			break;
